@@ -37,6 +37,7 @@
 #include <iconv.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
+#include <langinfo.h>
 
 #include "libvhd.h"
 #include "relative-path.h"
@@ -1296,6 +1297,7 @@ vhd_macx_encode_location(char *name, char **out, int *outlen)
 	size_t ibl, obl;
 	char *uri, *uri_utf8, *uri_utf8p, *ret;
 	const char *urip;
+	char *codeset;
 
 	err     = 0;
 	ret     = NULL;
@@ -1304,7 +1306,7 @@ vhd_macx_encode_location(char *name, char **out, int *outlen)
 	len     = strlen(name) + strlen("file://");
 
 	ibl     = len;
-	obl     = len;
+	obl     = len * 2;
 
 	urip = uri = malloc(ibl + 1);
 	uri_utf8 = uri_utf8p = malloc(obl);
@@ -1312,7 +1314,8 @@ vhd_macx_encode_location(char *name, char **out, int *outlen)
 	if (!uri || !uri_utf8)
 		return -ENOMEM;
 
-	cd = iconv_open("UTF-8", "ASCII");
+	codeset = nl_langinfo(CODESET);
+	cd = iconv_open("UTF-8", codeset);
 	if (cd == (iconv_t)-1) {
 		err = -errno;
 		goto out;
@@ -1325,7 +1328,7 @@ vhd_macx_encode_location(char *name, char **out, int *outlen)
 	    (char **)
 #endif
 	    &urip, &ibl, &uri_utf8p, &obl) == (size_t)-1 ||
-	    ibl || obl) {
+	    ibl) {
 		err = (errno ? -errno : -EIO);
 		goto out;
 	}
@@ -1357,6 +1360,7 @@ vhd_w2u_encode_location(char *name, char **out, int *outlen)
 	size_t ibl, obl;
 	char *uri, *uri_utf16, *uri_utf16p, *tmp, *ret;
 	const char *urip;
+	char *codeset;
 
 	err     = 0;
 	ret     = NULL;
@@ -1404,7 +1408,8 @@ vhd_w2u_encode_location(char *name, char **out, int *outlen)
 	 * MICROSOFT_COMPAT
 	 * little endian unicode here 
 	 */
-	cd = iconv_open("UTF-16LE", "ASCII");
+	codeset = nl_langinfo(CODESET);
+	cd = iconv_open("UTF-16LE", codeset);
 	if (cd == (iconv_t)-1) {
 		err = -errno;
 		goto out;
@@ -1415,7 +1420,7 @@ vhd_w2u_encode_location(char *name, char **out, int *outlen)
 	    (char **)
 #endif
 	    &urip, &ibl, &uri_utf16p, &obl) == (size_t)-1 ||
-	    ibl || obl) {
+	    ibl) {
 		err = (errno ? -errno : -EIO);
 		goto out;
 	}
@@ -1447,11 +1452,13 @@ vhd_macx_decode_location(const char *in, char *out, int len)
 	iconv_t cd;
 	char *name;
 	size_t ibl, obl;
+	char *codeset;
 
 	name = out;
 	ibl  = obl = len;
 
-	cd = iconv_open("ASCII", "UTF-8");
+	codeset = nl_langinfo(CODESET);
+	cd = iconv_open(codeset, "UTF-8");
 	if (cd == (iconv_t)-1) 
 		return NULL;
 
@@ -1479,11 +1486,13 @@ vhd_w2u_decode_location(const char *in, char *out, int len, char *utf_type)
 	iconv_t cd;
 	char *name, *tmp;
 	size_t ibl, obl;
+	char *codeset;
 
 	tmp = name = out;
 	ibl = obl  = len;
 
-	cd = iconv_open("ASCII", utf_type);
+	codeset = nl_langinfo(CODESET);
+	cd = iconv_open(codeset, utf_type);
 	if (cd == (iconv_t)-1) 
 		return NULL;
 
@@ -2450,6 +2459,7 @@ vhd_initialize_header_parent_name(vhd_context_t *ctx, const char *parent_path)
 	size_t ibl, obl;
 	char *ppath, *dst;
 	const char *pname;
+	char *codeset;
 
 	err   = 0;
 	pname = NULL;
@@ -2459,7 +2469,8 @@ vhd_initialize_header_parent_name(vhd_context_t *ctx, const char *parent_path)
 	 * MICROSOFT_COMPAT
 	 * big endian unicode here 
 	 */
-	cd = iconv_open(UTF_16BE, "ASCII");
+	codeset = nl_langinfo(CODESET);
+	cd = iconv_open(UTF_16BE, codeset);
 	if (cd == (iconv_t)-1) {
 		err = -errno;
 		goto out;
