@@ -2,6 +2,7 @@
 #include <xen/errno.h>
 #include <xen/init.h>
 #include <xen/lib.h>
+#include <asm/page.h>
 
 #ifndef efi_enabled
 const bool_t efi_enabled = 0;
@@ -9,7 +10,9 @@ const bool_t efi_enabled = 0;
 
 void __init efi_init_memory(void) { }
 
-paddr_t efi_rs_page_table(void)
+void efi_update_l4_pgtable(unsigned int l4idx, l4_pgentry_t l4e) { }
+
+bool_t efi_rs_using_pgtables(void)
 {
     BUG();
     return 0;
