@@ -19,6 +19,7 @@
 #include <xen/ctype.h>
 #include <xen/symbols.h>
 #include <xen/lib.h>
+#include <xen/sched.h>
 #include <asm/div64.h>
 #include <asm/page.h>
 
@@ -191,45 +192,45 @@ static char *number(
     size -= precision;
     if (!(type&(ZEROPAD+LEFT))) {
         while(size-->0) {
-            if (buf <= end)
+            if (buf < end)
                 *buf = ' ';
             ++buf;
         }
     }
     if (sign) {
-        if (buf <= end)
+        if (buf < end)
             *buf = sign;
         ++buf;
     }
     if (type & SPECIAL) {
-        if (buf <= end)
+        if (buf < end)
             *buf = '0';
         ++buf;
         if (base == 16) {
-            if (buf <= end)
+            if (buf < end)
                 *buf = digits[33];
             ++buf;
         }
     }
     if (!(type & LEFT)) {
         while (size-- > 0) {
-            if (buf <= end)
+            if (buf < end)
                 *buf = c;
             ++buf;
         }
     }
     while (i < precision--) {
-        if (buf <= end)
+        if (buf < end)
             *buf = '0';
         ++buf;
     }
     while (i-- > 0) {
-        if (buf <= end)
+        if (buf < end)
             *buf = tmp[i];
         ++buf;
     }
     while (size-- > 0) {
-        if (buf <= end)
+        if (buf < end)
             *buf = ' ';
         ++buf;
     }
@@ -239,22 +240,22 @@ static char *number(
 static char *string(char *str, char *end, const char *s,
                     int field_width, int precision, int flags)
 {
-    int i, len = strnlen(s, precision);
+    int i, len = (precision < 0) ? strlen(s) : strnlen(s, precision);
 
     if (!(flags & LEFT)) {
         while (len < field_width--) {
-            if (str <= end)
+            if (str < end)
                 *str = ' ';
             ++str;
         }
     }
     for (i = 0; i < len; ++i) {
-        if (str <= end)
+        if (str < end)
             *str = *s;
         ++str; ++s;
     }
     while (len < field_width--) {
-        if (str <= end)
+        if (str < end)
             *str = ' ';
         ++str;
     }
@@ -271,6 +272,34 @@ static char *pointer(char *str, char *end, const char **fmt_ptr,
     /* Custom %p suffixes. See XEN_ROOT/docs/misc/printk-formats.txt */
     switch ( fmt[1] )
     {
+    case 'h': /* Raw buffer as hex string. */
+    {
+        const uint8_t *hex_buffer = arg;
+        unsigned int i;
+
+        /* Consumed 'h' from the format string. */
+        ++*fmt_ptr;
+
+        /* Bound user count from %* to between 0 and 64 bytes. */
+        if ( field_width <= 0 )
+            return str;
+        if ( field_width > 64 )
+            field_width = 64;
+
+        for ( i = 0; ; )
+        {
+            /* Each byte: 2 chars, 0-padded, base 16, no hex prefix. */
+            str = number(str, end, hex_buffer[i], 16, 2, -1, ZEROPAD);
+
+            if ( ++i == field_width )
+                return str;
+
+            if ( str < end )
+                *str = ' ';
+            ++str;
+        }
+    }
+
     case 's': /* Symbol name with offset and size (iff offset != 0) */
     case 'S': /* Symbol name unconditionally with offset and size */
     {
@@ -293,7 +322,7 @@ static char *pointer(char *str, char *end, const char **fmt_ptr,
         {
             /* Print '+<offset>/<len>' */
             str = number(str, end, sym_offset, 16, -1, -1, SPECIAL|SIGN|PLUS);
-            if ( str <= end )
+            if ( str < end )
                 *str = '/';
             ++str;
             str = number(str, end, sym_size, 16, -1, -1, SPECIAL);
@@ -301,6 +330,24 @@ static char *pointer(char *str, char *end, const char **fmt_ptr,
 
         return str;
     }
+
+    case 'v': /* d<domain-id>v<vcpu-id> from a struct vcpu */
+    {
+        const struct vcpu *v = arg;
+
+        ++*fmt_ptr;
+        if ( unlikely(v->domain->domain_id == DOMID_IDLE) )
+            str = string(str, end, "IDLE", -1, -1, 0);
+        else
+        {
+            if ( str < end )
+                *str = 'd';
+            str = number(str + 1, end, v->domain->domain_id, 10, -1, -1, 0);
+        }
+        if ( str < end )
+            *str = 'v';
+        return number(str + 1, end, v->vcpu_id, 10, -1, -1, 0);
+    }
     }
 
     if ( field_width == -1 )
@@ -351,16 +398,16 @@ int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
     BUG_ON(((int)size < 0) || ((unsigned int)size != size));
 
     str = buf;
-    end = buf + size - 1;
+    end = buf + size;
 
-    if (end < buf - 1) {
+    if (end < buf) {
         end = ((void *) -1);
-        size = end - buf + 1;
+        size = end - buf;
     }
 
     for (; *fmt ; ++fmt) {
         if (*fmt != '%') {
-            if (str <= end)
+            if (str < end)
                 *str = *fmt;
             ++str;
             continue;
@@ -426,17 +473,17 @@ int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
         case 'c':
             if (!(flags & LEFT)) {
                 while (--field_width > 0) {
-                    if (str <= end)
+                    if (str < end)
                         *str = ' ';
                     ++str;
                 }
             }
             c = (unsigned char) va_arg(args, int);
-            if (str <= end)
+            if (str < end)
                 *str = c;
             ++str;
             while (--field_width > 0) {
-                if (str <= end)
+                if (str < end)
                     *str = ' ';
                 ++str;
             }
@@ -471,7 +518,7 @@ int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
             continue;
 
         case '%':
-            if (str <= end)
+            if (str < end)
                 *str = '%';
             ++str;
             continue;
@@ -494,11 +541,11 @@ int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
             break;
 
         default:
-            if (str <= end)
+            if (str < end)
                 *str = '%';
             ++str;
             if (*fmt) {
-                if (str <= end)
+                if (str < end)
                     *str = *fmt;
                 ++str;
             } else {
@@ -527,11 +574,14 @@ int vsnprintf(char *buf, size_t size, const char *fmt, va_list args)
         str = number(str, end, num, base,
                      field_width, precision, flags);
     }
-    if (str <= end)
-        *str = '\0';
-    else if (size > 0)
-        /* don't write out a null byte if the buf size is zero */
-        *end = '\0';
+
+    /* don't write out a null byte if the buf size is zero */
+    if (size > 0) {
+        if (str < end)
+            *str = '\0';
+        else
+            end[-1] = '\0';
+    }
     /* the trailing null byte doesn't count towards the total
      * ++str;
      */
@@ -617,6 +667,60 @@ int scnprintf(char * buf, size_t size, const char *fmt, ...)
 }
 EXPORT_SYMBOL(scnprintf);
 
+/**
+ * vasprintf - Format a string and allocate a buffer to place it in
+ *
+ * @bufp: Pointer to a pointer to receive the allocated buffer
+ * @fmt: The format string to use
+ * @args: Arguments for the format string
+ *
+ * -ENOMEM is returned on failure and @bufp is not touched.
+ * On success, 0 is returned. The buffer passed back is
+ * guaranteed to be null terminated. The memory is allocated
+ * from xenheap, so the buffer should be freed with xfree().
+ */
+int vasprintf(char **bufp, const char *fmt, va_list args)
+{
+    va_list args_copy;
+    size_t size;
+    char *buf;
+
+    va_copy(args_copy, args);
+    size = vsnprintf(NULL, 0, fmt, args_copy);
+    va_end(args_copy);
+
+    buf = xmalloc_array(char, ++size);
+    if ( !buf )
+        return -ENOMEM;
+
+    (void) vsnprintf(buf, size, fmt, args);
+
+    *bufp = buf;
+    return 0;
+}
+
+/**
+ * asprintf - Format a string and place it in a buffer
+ * @bufp: Pointer to a pointer to receive the allocated buffer
+ * @fmt: The format string to use
+ * @...: Arguments for the format string
+ *
+ * -ENOMEM is returned on failure and @bufp is not touched.
+ * On success, 0 is returned. The buffer passed back is
+ * guaranteed to be null terminated. The memory is allocated
+ * from xenheap, so the buffer should be freed with xfree().
+ */
+int asprintf(char **bufp, const char *fmt, ...)
+{
+    va_list args;
+    int i;
+
+    va_start(args, fmt);
+    i=vasprintf(bufp,fmt,args);
+    va_end(args);
+    return i;
+}
+
 /*
  * Local variables:
  * mode: C
